import os
import shutil

import pytest
from papermill.engines import papermill_engines
from papermill.execute import raise_for_execution_errors
from papermill.iorw import load_notebook_node
from papermill.utils import nb_kernel_name

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
NBS_DIR = os.path.join(TEST_DIR, "nbs")
EXAMPLES_DIR = os.path.join(os.path.dirname(TEST_DIR), "examples")
NBS_FAIL_DIR = os.path.join(TEST_DIR, "nbs_fail")

shutil.rmtree(NBS_FAIL_DIR, ignore_errors=True)
os.mkdir(NBS_FAIL_DIR)


def _find_notebooks(dir):
    dir = os.path.abspath(os.path.expanduser(dir))
    for root, _, files in os.walk(dir):
        for file in files:
            if ("checkpoint" in file) or (not file.endswith(".ipynb")):
                continue
            yield os.path.join(root, file)


TEST_NOTEBOOKS = [
    *sorted(_find_notebooks(NBS_DIR)),
    *sorted(_find_notebooks(EXAMPLES_DIR)),
]


@pytest.mark.parametrize("path", sorted(TEST_NOTEBOOKS))
def test_nbs(path):
    fn = os.path.basename(path)
    nb = load_notebook_node(path)
    nb = papermill_engines.execute_notebook_with_engine(
        engine_name=None,
        nb=nb,
        kernel_name=nb_kernel_name(nb, None),
        input_path=path,
        output_path=None,
    )
    raise_for_execution_errors(nb, os.path.join(NBS_FAIL_DIR, fn))



""" FDE Lumerical Backend """

from typing import Optional

import numpy as np
from pydantic import validate_arguments
from pydantic.types import PositiveInt

from ..cross_section import CrossSection
from ..mode import Mode, normalize_energy, zero_phase

_global = {"sim": None}


def _assert_default_mesh_setting(condition, param_name):
    if condition:
        raise NotImplementedError(
            f"Setting mesh.{param_name} is currently not supported in the Lumerical Backend. "
            "Please open an issue of submit a PR on GitHub to fix this: ",
            "https://github.com/flaport/meow",
        )


def set_sim(
    sim: "lumapi.MODE",  # type: ignore
):
    _global["sim"] = sim


def get_sim():
    sim = _global["sim"]
    if sim is None:
        raise ValueError(
            "Could not start Lumerical simulation. "
            "Please either pass the `lumapi.MODE` simulation object as an argument to `compute_modes_lumerical` or "
            "use `set_sim(sim)` to globally set the lumapi.MODE simulation object."
        )
    return sim


# @validate_arguments
def compute_modes_lumerical(
    cs: CrossSection,
    num_modes: PositiveInt = 10,
    sim: Optional["lumapi.MODE"] = None,  # type: ignore
    unit: float = 1e-6,
):
    """compute ``Modes` for a given ``FdeSpec` (Lumerical backend)

    Args:
        sim: the lumerical simulation object
        spec: The FDE simulation specification
        unit: Conversion factor between MEOW unit (probably um) and Lumerical unit (probably m).
    """
    _assert_default_mesh_setting(cs.cell.mesh.angle_phi == 0, "angle_phi")
    _assert_default_mesh_setting(cs.cell.mesh.angle_theta == 0, "angle_theta")
    _assert_default_mesh_setting(cs.cell.mesh.bend_radius > 1e10, "bend_radius")

    if sim is None:
        sim = get_sim()

    assert sim is not None

    sim.switchtolayout()
    sim.deleteall()
    for s in cs.structures:
        s._lumadd(sim, cs.env, unit, "yzx")

    sim.select("FDE")
    sim.delete()
    sim.addfde(
        background_index=1.0,
        solver_type="2D X normal",
        x=float(cs.cell.z * unit),
        y_min=float(cs.mesh.x.min() * unit),
        y_max=float(cs.mesh.x.max() * unit),
        z_min=float(cs.mesh.y.min() * unit),
        z_max=float(cs.mesh.y.max() * unit),
        define_z_mesh_by="number of mesh cells",
        define_y_mesh_by="number of mesh cells",
        mesh_cells_y=cs.mesh.x_.shape[0],
        mesh_cells_z=cs.mesh.y_.shape[0],
    )
    sim.setanalysis("number of trial modes", int(num_modes))
    sim.setanalysis("search", "near n")
    sim.setanalysis("use max index", True)
    sim.setanalysis("wavelength", float(cs.env.wl * unit))
    sim.findmodes()
    modes = []
    for j in range(1, num_modes + 1):
        mode = Mode(
            neff=sim.getdata(f"mode{j}", "neff").ravel().item(),
            cs=cs,
            Ez=sim.getdata(f"mode{j}", "Ex").squeeze()[:-1, :-1],
            Ex=sim.getdata(f"mode{j}", "Ey").squeeze()[:-1, :-1],
            Ey=sim.getdata(f"mode{j}", "Ez").squeeze()[:-1, :-1],
            Hz=sim.getdata(f"mode{j}", "Hx").squeeze()[:-1, :-1],
            Hx=sim.getdata(f"mode{j}", "Hy").squeeze()[:-1, :-1],
            Hy=sim.getdata(f"mode{j}", "Hz").squeeze()[:-1, :-1],
        )
        mode = normalize_energy(mode)
        mode = zero_phase(mode)
        modes.append(mode)

    return sorted(modes, key=lambda m: np.real(m.neff), reverse=True)



""" An EigenMode """

import pickle
from itertools import product
from typing import List, Tuple, Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from pydantic import Field, PrivateAttr
from scipy.constants import epsilon_0 as eps0
from scipy.constants import mu_0 as mu0
from scipy.linalg import norm
from .integrate import integrate_2d

from .base_model import BaseModel
from .cross_section import CrossSection


class Mode(BaseModel):
    """A `Mode` contains the field information for a given `CrossSection`."""

    neff: complex = Field(description="the effective index of the mode")
    cs: CrossSection = Field(
        description="the index cross section for which the mode was calculated"
    )
    Ex: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Ex-fields of the mode"
    )
    Ey: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Ey-fields of the mode"
    )
    Ez: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Ez-fields of the mode"
    )
    Hx: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Hx-fields of the mode"
    )
    Hy: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Hy-fields of the mode"
    )
    Hz: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Hz-fields of the mode"
    )

    _Px: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = PrivateAttr()
    _Py: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = PrivateAttr()
    _Pz: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = PrivateAttr()
    _A: np.float_ = PrivateAttr()

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._Px = None
        self._Py = None
        self._Pz = None
        self._A = None

    @property
    def te_fraction(self):
        """the TE polarization fraction of the mode."""
        return te_fraction(self)

    def _calc_poynting(self):
        """calculate and cache the poynting vector"""
        vecE = np.stack([self.Ex, self.Ey, self.Ez], axis=-1)
        vecH = np.stack([self.Hx, self.Hy, self.Hz], axis=-1)
        vecP = np.cross(vecE, vecH)
        self._Px, self._Py, self._Pz = np.rollaxis(vecP, -1)

    def _calc_area(self):
        vecE = np.stack([self.Ex, self.Ey, self.Ez], axis=-1)
        E_sq = norm(vecE, axis=-1, ord=2)
        E_qu = E_sq**2
        x = self.cs.mesh.x_
        y = self.cs.mesh.y_
        self._A = integrate_2d(x, y, E_sq) ** 2 / integrate_2d(x, y, E_qu)

    @property
    def Px(self):
        if self._Px is None:
            self._calc_poynting()
        return self._Px

    @property
    def Py(self):
        if self._Py is None:
            self._calc_poynting()
        return self._Py

    @property
    def Pz(self):
        if self._Pz is None:
            self._calc_poynting()
        return self._Pz

    @property
    def A(self):
        """mode area"""
        if self._A is None:
            self._calc_area()
        return self._A

    @property
    def env(self):
        return self.cs.env

    @property
    def mesh(self):
        return self.cs.mesh

    @property
    def cell(self):
        return self.cs.cell

    def _visualize(
        self,
        title=None,
        fields=None,
        ax=None,
        n_cmap=None,
        mode_cmap=None,
        num_levels=8,
        operation=lambda x: np.abs(x) ** 2,
    ):
        if fields is None or len(fields) == 0:
            fields = ["Ex"]

        if len(fields) > 1:
            if ax is None:
                _, ax = plt.subplots(len(fields), 1, figsize=(5, len(fields) * 3))
            if len(ax) < len(fields):
                raise ValueError(
                    f"Not enough axes supplied for the number of fields "
                    f"to plot! {len(ax)} < {len(fields)}."
                )
            for field, ax_ in zip(fields, ax):
                self._visualize(
                    field,
                    ax=ax_,
                    n_cmap=n_cmap,
                    mode_cmap=mode_cmap,
                    num_levels=num_levels,
                )
            return

        field = fields[0]
        valid_fields = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz", "Px", "Py", "Pz"]
        if field not in valid_fields:
            raise ValueError(
                f"Invalid field {field!r}. Valid fields: {', '.join(valid_fields)}."
            )

        if ax is None:
            ax = plt.gca()
            plt.axis("scaled")
        plt.sca(ax)

        x, y = "x", "y"  # currently only propagation in z supported, see Mesh2d
        c = field[-1]
        if n_cmap is None:
            n_cmap = colors.LinearSegmentedColormap.from_list(
                name="c_cmap", colors=["#ffffff", "#c1d9ed"]
            )
        if mode_cmap is None:
            mode_cmap = "inferno"
        X = getattr(self.mesh, f"X{c}")
        Y = getattr(self.mesh, f"Y{c}")
        mode = operation(getattr(self, field))
        plt.contour(X, Y, mode, cmap=mode_cmap, levels=np.linspace(mode.min(), mode.max(), num_levels))  # fmt: skip
        plt.colorbar(label="mode")

        n = np.real(getattr(self.cs, f"n{c}"))
        plt.pcolormesh(X, Y, n, cmap=n_cmap)
        plt.colorbar(label="n")
        plt.xlabel(x)
        plt.ylabel(y)
        plt.grid(True, alpha=0.4)
        if title is None:
            plt.title(f"{field} [neff={float(np.real(self.neff)):.6f}]")
        else:
            plt.title(title)
        plt.xlim(X.min(), X.max())
        plt.ylim(Y.min(), Y.max())
        plt.show()

    def save(self, filename):
        with open(filename, "wb") as file:
            pickle.dump(self, file)

    @classmethod
    def load(cls, filename):
        with open(filename, "rb") as file:
            return pickle.load(file)


Modes = List[Mode]


def zero_phase(mode: Mode) -> Mode:
    """normalize (zero out) the phase of a `Mode`"""
    e = np.abs(energy_density(mode))
    m, n = np.array(np.where(e == e.max()))[:, 0]
    phase = np.exp(-1j * np.angle(np.array(mode.Hx))[m][n])
    new_mode = Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=mode.Ex * phase,
        Ey=mode.Ey * phase,
        Ez=mode.Ez * phase,
        Hx=mode.Hx * phase,
        Hy=mode.Hy * phase,
        Hz=mode.Hz * phase,
    )
    if _sum_around(np.real(new_mode.Hx), m, n) < 0:
        new_mode = invert_mode(new_mode)
    return new_mode


def _sum_around(field, m, n, r=2):
    total = 0
    idxs = range(-r, r + 1)
    idx_tups = product(idxs, idxs)
    M, N = field.shape
    for i, j in idx_tups:
        m_ = min(m + i, M - 1) if i >= 0 else max(m + i, 0)
        n_ = min(n + j, N - 1) if i >= 0 else max(n + j, 0)
        total = total + field[m_, n_]
    return total


def invert_mode(mode: Mode) -> Mode:
    """invert a `Mode`"""
    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=-mode.Ex,
        Ey=-mode.Ey,
        Ez=-mode.Ez,
        Hx=-mode.Hx,
        Hy=-mode.Hy,
        Hz=-mode.Hz,
    )


def inner_product(mode1: Mode, mode2: Mode) -> float:
    """the inner product of a `Mode` with another `Mode` is uniquely defined."""
    mesh = mode1.mesh
    cross = mode1.Ex * mode2.Hy.conj() - mode1.Ey * mode2.Hx.conj()
    return 0.25 * np.trapz(np.trapz(cross, mesh.y_), mesh.x_)


def normalize_product(mode: Mode) -> Mode:
    """normalize a `Mode` according to the `inner_product` with itself"""
    factor = np.sqrt(inner_product(mode, mode))
    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=mode.Ex / factor,
        Ey=mode.Ey / factor,
        Ez=mode.Ez / factor,
        Hx=mode.Hx / factor,
        Hy=mode.Hy / factor,
        Hz=mode.Hz / factor,
    )


def electric_energy_density(
    mode: Mode,
) -> np.ndarray[Tuple[int, int], np.dtype[np.float_]]:
    """get the electric energy density contained in a `Mode`"""
    epsx, epsy, epsz = mode.cs.nx**2, mode.cs.ny**2, mode.cs.nz**2
    return (
        0.5
        * eps0
        * (
            epsx * np.abs(mode.Ex) ** 2
            + epsy * np.abs(mode.Ey) ** 2
            + epsz * np.abs(mode.Ez) ** 2
        )
    )


def electric_energy(mode: Mode) -> float:
    """get the electric energy contained in a `Mode`"""
    return electric_energy_density(mode).sum()


def magnetic_energy_density(
    mode: Mode,
) -> np.ndarray[Tuple[int, int], np.dtype[np.float_]]:
    """get the magnetic energy density contained in a `Mode`"""
    return (
        0.5 * mu0 * (np.abs(mode.Hx) ** 2 + np.abs(mode.Hy) ** 2 + np.abs(mode.Hz) ** 2)
    )


def magnetic_energy(mode: Mode) -> float:
    """get the magnetic energy contained in a `Mode`"""
    return magnetic_energy_density(mode).sum()


def energy_density(mode: Mode) -> np.ndarray[Tuple[int, int], np.dtype[np.float_]]:
    """get the energy density contained in a `Mode`"""
    return electric_energy_density(mode) + magnetic_energy_density(mode)


def energy(mode: Mode) -> float:
    """get the energy contained in a `Mode`"""
    return energy_density(mode).sum()


def normalize_energy(mode: Mode) -> Mode:
    """normalize a mode according to the energy it contains"""
    e = np.sqrt(2 * electric_energy(mode))
    h = np.sqrt(2 * magnetic_energy(mode))

    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=mode.Ex / e,
        Ey=mode.Ey / e,
        Ez=mode.Ez / e,
        Hx=mode.Hx / h,
        Hy=mode.Hy / h,
        Hz=mode.Hz / h,
    )


def te_fraction(mode: Mode) -> float:
    """the TE polarization fraction of the `Mode`"""
    epsx = mode.cs.nx**2
    e = np.sum(0.5 * eps0 * epsx * np.abs(mode.Ex) ** 2)
    h = np.sum(0.5 * mu0 * np.abs(mode.Hx) ** 2)
    return float(e / (e + h))



""" an EME Cell """

from typing import Any, Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap, to_rgba
from pydantic import Field

from .base_model import BaseModel
from .mesh import Mesh2d
from .structures import Structure, sort_structures


class Cell(BaseModel):
    """A `Cell` defines a location in a `Structure` associated with a `Mesh`"""

    structures: List[Structure] = Field(
        description="the structures which will be sliced by the cell"
    )
    mesh: Mesh2d = Field(description="the mesh to slice the structures with")
    z_min: float = Field(description="the starting z-coordinate of the cell")
    z_max: float = Field(description="the ending z-coordinate of the cell")
    extra: Dict[str, Any] = Field(
        default_factory=lambda: {}, description="extra metadata"
    )

    @property
    def mx(self):
        """(derived) the material cross section at the Ex grid (integer y-coords, half-integer x-coords)"""
        return self.extra["mx"]

    @property
    def my(self):
        """(derived) the material cross section at the Ey grid (half-integer y-coords, integer x-coords)"""
        return self.extra["my"]

    @property
    def mz(self):
        """(derived) the material cross section at the Ez grid (integer y-coords, integer x-coords)"""
        return self.extra["mz"]

    @property
    def materials(self):
        """(derived) the materials in the cell"""
        return self.extra["materials"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        structures = sort_structures(self.structures)
        mx, my, mz = [np.zeros(self.mesh.Xx.shape, dtype=int) for _ in range(3)]
        z = 0.5 * (self.z_min + self.z_max)
        # TODO: ideally we should downselect the relevant structures here...
        # structures not at z-location should ideally be ignored.
        materials = []
        for structure in structures:
            mask_x, mask_y, mask_z = structure.geometry._mask2d(self.mesh, z)
            if (not mask_x.any()) or (not mask_y.any()) or (not mask_z.any()):
                continue
            try:
                material_index = materials.index(structure.material) + 1
            except ValueError:
                materials.append(structure.material)
                material_index = len(materials)
            mx[mask_x] = material_index
            my[mask_y] = material_index
            mz[mask_z] = material_index
        self.extra["mx"] = mx
        self.extra["my"] = my
        self.extra["mz"] = mz
        self.extra["materials"] = materials

    @property
    def z(self):
        return 0.5 * (self.z_min + self.z_max)

    @property
    def length(self):
        return np.abs(self.z_max - self.z_min)

    def _visualize(self, c="z", axs=None):
        colors = [(0, 0, 0, 0)] + [
            to_rgba(m.meta.get("color", (0, 0, 0, 0))) for m in self.materials
        ]
        cmap = ListedColormap(colors=colors)  # type: ignore
        if axs is None:
            _, axs = plt.subplots(1, len(c), figsize=(3 * len(c), 3))
        c_list = list(c)
        if any(c not in "xyz" for c in c_list):
            raise ValueError(f"Invalid component. Got: {c}. Should be 'x', 'y' or 'z'.")
        axs = np.array(axs, dtype=object).ravel()
        for ax, c in zip(axs, c_list):
            X = getattr(self.mesh, f"X{c}")
            Y = getattr(self.mesh, f"Y{c}")
            m = getattr(self, f"m{c}")
            plt.sca(ax)
            if len(c_list) > 1:
                plt.title(f"m{c}")
            plt.pcolormesh(X, Y, m, cmap=cmap, vmin=0, vmax=len(self.materials))
            plt.axis("scaled")
            plt.grid(True)

    class Config:
        fields = {"extra": {"exclude": True}}


def create_cells(
    structures: List[Structure],
    mesh: Union[Mesh2d, List[Mesh2d]],
    Ls: np.ndarray[Tuple[int], np.dtype[np.float_]],
    z_min: float = 0.0,
) -> List[Cell]:
    """easily create multiple `Cell` objects given a `Mesh` and a collection of cell lengths"""

    Ls = np.asarray(Ls, float)
    if Ls.ndim != 1:
        raise ValueError(f"Ls should be 1D. Got shape: {Ls.shape}.")
    if Ls.shape[0] < 0:
        raise ValueError(f"length of Ls array should be nonzero. Got: {Ls}.")

    meshes = [mesh] * Ls.shape[0] if isinstance(mesh, Mesh2d) else mesh
    if len(Ls) != len(meshes):
        raise ValueError(
            f"Number of meshes should correspond to number of lengths (length of Ls). Got {len(meshes)} != {len(Ls)}."
        )

    z = np.cumsum(np.concatenate([np.asarray([z_min], float), Ls]))
    cells = [
        Cell(structures=structures, mesh=mesh, z_min=z_min, z_max=z_max)
        for mesh, (z_min, z_max) in zip(meshes, zip(z[:-1], z[1:]))
    ]

    return cells



""" a 2D Mesh """

from typing import Literal, Tuple

import numpy as np
from pydantic import Field
from pydantic.types import NonNegativeInt

from .base_model import BaseModel


class Mesh(BaseModel):
    """[BaseClass] a ``Mesh`` describes how a ``Structure`` is discretized"""


class Mesh2d(Mesh):
    """a 2D Mesh or ``Mesh2D`` describes how a ``Structure`` is discritized into a ``Cell`` or ``CrossSection``"""

    x: np.ndarray[Tuple[int], np.dtype[np.float_]] = Field(
        description="x-coordinates of the mesh (Ez locations, i.e. corners of the 2D cell)"
    )
    y: np.ndarray[Tuple[int], np.dtype[np.float_]] = Field(
        description="y-coordinates of the mesh (Ez locations, i.e. corners of the 2D cell)"
    )
    angle_phi: float = Field(
        default=0.0,
        description="Azimuth angle of the propagation axis in the plane orthogonal to the mesh.",
    )
    angle_theta: float = Field(
        default=0.0,
        description="Polar angle of the propagation axis from the injection axis.",
    )
    bend_radius: float = Field(
        default=np.inf,
        description=(
            "A curvature radius for simulation of waveguide bends. "
            "Tidy3D: Can be negative, in which case the mode plane center has a smaller value than "
            "the curvature center along the tangential axis perpendicular to the bend axis."
        ),
    )
    bend_axis: Literal[0, 1] = Field(
        default=1,
        description=(
            "Index into the two tangential axes defining the normal to the plane in which the bend lies. "
            "This must be provided if ``bend_radius`` is not ``None``. For example, for a ring in the "
            "global xy-plane, and a mode plane in either the xz or the yz plane, the ``bend_axis`` is "
            "always 1 (the global z axis)."
        ),
    )
    num_pml: Tuple[NonNegativeInt, NonNegativeInt] = Field(
        default=(0, 0),
        description="Number of standard pml layers to add in the two tangential axes.",
    )

    @property
    def dx(self):
        """x-coordinate mesh step (Hz locations, i.e. center of the 2D cell)"""
        return self.x[1:] - self.x[:-1]

    @property
    def dy(self):
        """y-coordinate mesh step (Hz locations, i.e. center of the 2D cell)"""
        return self.y[1:] - self.y[:-1]

    @property
    def x_(self):
        """x-coordinates of the mesh (Hz locations, i.e. center of the 2D cell)"""
        return 0.5 * (self.x[1:] + self.x[:-1])

    @property
    def y_(self):
        """y-coordinates of the mesh (Hz locations, i.e. center of the 2D cell)"""
        return 0.5 * (self.y[1:] + self.y[:-1])

    @property
    def X(self):
        return np.meshgrid(self.y_, self.x_)[1]

    @property
    def Y(self):
        return np.meshgrid(self.y_, self.x_)[0]

    @property
    def Xx(self):
        return np.meshgrid(self.y[:-1], self.x_)[1]

    @property
    def Yx(self):
        return np.meshgrid(self.y[:-1], self.x_)[0]

    @property
    def Xy(self):
        return np.meshgrid(self.y_, self.x[:-1])[1]

    @property
    def Yy(self):
        return np.meshgrid(self.y_, self.x[:-1])[0]

    @property
    def Xz(self):
        return np.meshgrid(self.y[:-1], self.x[:-1])[1]

    @property
    def Yz(self):
        return np.meshgrid(self.y[:-1], self.x[:-1])[0]

    def __eq__(self, other):
        try:
            x_eq = ((self.x - other.x) < 1e-6).all()
            y_eq = ((self.y - other.y) < 1e-6).all()
        except Exception:
            return False
        return x_eq and y_eq



from functools import lru_cache, wraps
from typing import List

import meow as mw
import numpy as np
from pydantic import Field, parse_raw_as


class SimpleMode(mw.BaseModel):
    neff: complex = Field(description="the effective index of the mode")
    cs: mw.CrossSection = Field(
        description="the index cross section for which the mode was calculated"
    )
    te_fraction: float = Field(description="the TE polarization fraction of the mode.")

    @property
    def env(self):
        return self.cs.env

    @property
    def mesh(self):
        return self.cs.mesh

    @property
    def cell(self):
        return self.cs.cell


SIMPLE_MODE_DATA = {
    "neff": {"real": 3.526228477887, "imag": -0.03995016558},
    "cs": {
        "cell": {
            "structures": [
                {
                    "material": {
                        "name": "SiO2",
                        "params": {"wl": [1.2]},
                        "n": {"real": [1.955], "imag": [0.0]},
                    },
                    "geometry": {
                        "type": "Box",
                        "x_min": -1.0,
                        "x_max": 1.0,
                        "y_min": 0.0,
                        "y_max": 3.5,
                        "z_min": -1.0,
                        "z_max": 1.0,
                    },
                    "mesh_order": 5,
                },
                {
                    "material": {
                        "name": "SiO2",
                        "params": {"wl": [1.2]},
                        "n": {"real": [1.955], "imag": [0.0]},
                    },
                    "geometry": {
                        "type": "Box",
                        "x_min": -4.0,
                        "x_max": 4.0,
                        "y_min": -2.0,
                        "y_max": 2.3,
                        "z_min": -1.0,
                        "z_max": 1.0,
                    },
                    "mesh_order": 5,
                },
                {
                    "material": {
                        "name": "Si",
                        "params": {"wl": [1.2], "T": [38.0]},
                        "n": {"real": [3.526617910662], "imag": [0.0]},
                    },
                    "geometry": {
                        "type": "Box",
                        "x_min": -0.75,
                        "x_max": 0.75,
                        "y_min": 0.0,
                        "y_max": 3.0,
                        "z_min": -1.0,
                        "z_max": 1.0,
                    },
                    "mesh_order": 5,
                },
                {
                    "material": {
                        "name": "Si",
                        "params": {"wl": [1.2], "T": [38.0]},
                        "n": {"real": [3.526617910662], "imag": [0.0]},
                    },
                    "geometry": {
                        "type": "Box",
                        "x_min": -4.0,
                        "x_max": 4.0,
                        "y_min": 0.0,
                        "y_max": 1.8,
                        "z_min": -1.0,
                        "z_max": 1.0,
                    },
                    "mesh_order": 5,
                },
            ],
            "mesh": {
                "x": [],
                "y": [],
                "angle_phi": 0.0,
                "angle_theta": 0.0,
                "bend_radius": 6000.0,
                "bend_axis": 1,
                "num_pml": [6, 6],
            },
            "z_min": -1.0,
            "z_max": -1.0,
        },
        "env": {"wl": 1.2, "T": 38.0},
    },
    "te_fraction": 0.9998199906370098,
}


def test_deserialization():
    simple_mode = SimpleMode.parse_obj(SIMPLE_MODE_DATA)
    assert np.round(np.real(simple_mode.neff), 6) == np.round(
        SIMPLE_MODE_DATA["neff"]["real"], 6
    )
    assert np.round(np.imag(simple_mode.neff), 6) == np.round(
        SIMPLE_MODE_DATA["neff"]["imag"], 6
    )



""" MEOW: Modeling of Eigenmodes and Overlaps in Waveguides """

__author__ = "Floris Laporte"
__version__ = "0.2.0"

import warnings

# Silence Excessive Logging...

try:
    from loguru import logger

    logger.disable("gdsfactory")
except ImportError:
    pass

try:
    from numexpr.utils import log

    log.setLevel("CRITICAL")
except ImportError:
    pass

try:
    from rich import pretty

    old_install = pretty.install
    pretty.install = lambda *_, **__: None
    import tidy3d

    pretty.install = old_install
except ImportError:
    pass


try:
    import sax

    warnings.filterwarnings(action="ignore", module="sax")
except ImportError:
    pass

# from . import cell as cell
from . import base_model as base_model
from . import cross_section as cross_section
from . import eme as eme
from . import environment as environment
from . import fde as fde
from . import gds_structures as gds_structures
from . import geometries as geometries
from . import materials as materials
from . import mesh as mesh
from . import mode as mode
from . import structures as structures

# from . import visualize as visualize
from .cell import *
from .cross_section import *
from .eme import *
from .environment import *
from .fde import *
from .gds_structures import *
from .geometries import *
from .materials import *
from .mesh import *
from .mode import *
from .structures import *
from .visualize import *



""" FDE Tidy3d backend (default backend for MEOW) """

from types import SimpleNamespace
from typing import Optional

import numpy as np
from pydantic import validate_arguments
from pydantic.types import PositiveFloat, PositiveInt
from scipy.constants import c
from tidy3d.plugins.mode.solver import compute_modes as _compute_modes

from ..cross_section import CrossSection
from ..mode import Mode, Modes, normalize_energy, zero_phase


@validate_arguments
def compute_modes_tidy3d(
    cs: CrossSection,
    num_modes: PositiveInt = 10,
    target_neff: Optional[PositiveFloat] = None,
) -> Modes:
    """compute ``Modes`` for a given ``FdeSpec`` (Tidy3D backend)

    Args:
        cs: The ``CrossSection`` to calculate the modes for
        num_modes: Number of modes returned by mode solver.
        target_neff: Guess for initial effective index of the mode.
    """

    if num_modes < 1:
        raise ValueError("You need to request at least 1 mode.")

    bend_radius = None if cs.cell.mesh.bend_radius > 1e10 else cs.cell.mesh.bend_radius
    bend_axis = None if bend_radius is None else cs.cell.mesh.bend_axis

    ((Ex, Ey, Ez), (Hx, Hy, Hz)), neffs = (
        x.squeeze()
        for x in _compute_modes(
            eps_cross=[cs.nx**2, cs.ny**2, cs.nz**2],
            coords=[cs.mesh.x, cs.mesh.y],
            freq=c / (cs.env.wl * 1e-6),
            mode_spec=SimpleNamespace(
                num_modes=num_modes,
                angle_theta=cs.cell.mesh.angle_theta,
                angle_phi=cs.cell.mesh.angle_phi,
                bend_radius=bend_radius,
                bend_axis=bend_axis,
                target_neff=target_neff or cs.nx.max(),
                num_pml=cs.cell.mesh.num_pml,
                sort_by="largest_neff",
                filter_pol=None,
                precision="double",
            ),
        )
    )

    if num_modes == 1:
        modes = [
            Mode(
                cs=cs,
                Ex=Ex,
                Ey=Ey,
                Ez=Ez,
                Hx=Hx,
                Hy=Hy,
                Hz=Hz,
                neff=float(neffs.real) + 1j * float(neffs.imag),
            )
            for _ in range(num_modes)
        ]
    else:  # num_modes > 1
        modes = [
            Mode(
                cs=cs,
                Ex=Ex[..., i],
                Ey=Ey[..., i],
                Ez=Ez[..., i],
                Hx=Hx[..., i],
                Hy=Hy[..., i],
                Hz=Hz[..., i],
                neff=neffs[i],
            )
            for i in range(num_modes)
        ]

    modes = [zero_phase(normalize_energy(mode)) for mode in modes]
    modes = sorted(modes, key=lambda m: float(np.real(m.neff)), reverse=True)

    return modes



""" Geometries """

from secrets import token_hex
from typing import Dict, List, Literal, Tuple, Union, cast

import numpy as np
import shapely.geometry as sg
from matplotlib.colors import to_rgba as _to_rgba_mpl
from pydantic import Field, validator
from trimesh import Trimesh
from trimesh.creation import extrude_polygon
from trimesh.scene import Scene
from trimesh.transformations import rotation_matrix

from .base_model import BaseModel
from .mesh import Mesh2d

GEOMETRIES: Dict[str, type] = {}


class Geometry(BaseModel):
    type: str = ""

    def __init_subclass__(cls):
        GEOMETRIES[cls.__name__] = cls

    def __new__(cls, **kwargs):
        cls = GEOMETRIES.get(kwargs.get("type", cls.__name__), cls)
        return BaseModel.__new__(
            cls
        )  # , **kwargs) # weird pydantic black magic: kwargs are not explicitly propagated

    @validator("type", pre=True, always=True)
    def validate_type(cls, value):
        if not value:
            value = getattr(cls, "__name__", "Geometry")
        if value not in GEOMETRIES:
            raise ValueError(
                f"Invalid Geometry type. Got: {value!r}. Valid types: {GEOMETRIES}."
            )
        return value

    def _visualize(self, scale=None):
        scene = Scene()
        scene.add_geometry(self._trimesh(scale=scale))
        scene.apply_transform(rotation_matrix(-np.pi / 6, (0, 1, 0)))
        return scene.show()

    def _mask2d_single(self, X, Y, z):
        raise NotImplementedError(f"{self.__class__.__name__!r} cannot be masked.")

    def _mask2d(self, mesh: Mesh2d, z: float):
        mx = self._mask2d_single(mesh.Xx, mesh.Yx, z)
        my = self._mask2d_single(mesh.Xy, mesh.Yy, z)
        mz = self._mask2d_single(mesh.Xz, mesh.Yz, z)
        return mx, my, mz

    def _lumadd(self, sim, material_name, mesh_order, unit=1e-6, xyz="yzx"):
        raise NotImplementedError(
            f"{self.__class__.__name__!r} cannot be added to Lumerical."
        )

    def _trimesh(self, **kwargs):
        raise NotImplementedError(f"{self.__class__.__name__!r} cannot be visualized.")


Geometries = List[Geometry]


class Box(Geometry):
    """A Box is a simple rectangular cuboid"""

    x_min: float = Field(description="the minimum x-value of the box")
    x_max: float = Field(description="the maximum x-value of the box")
    y_min: float = Field(description="the minimum y-value of the box")
    y_max: float = Field(description="the maximum y-value of the box")
    z_min: float = Field(description="the minimum z-value of the box")
    z_max: float = Field(description="the maximum z-value of the box")

    def _mask2d_single(self, X, Y, z):
        if (z < self.z_min) or (self.z_max < z):
            return np.zeros_like(X, dtype=bool)
        return (
            (self.x_min <= X)
            & (X <= self.x_max)
            & (self.y_min <= Y)
            & (Y <= self.y_max)
        )

    def _lumadd(self, sim, material_name, mesh_order, unit, xyz):
        x, y, z = xyz
        name = token_hex(4)
        kwargs = {
            f"{x}_min": float(self.x_min * unit),
            f"{x}_max": float(self.x_max * unit),
            f"{y}_min": float(self.y_min * unit),
            f"{y}_max": float(self.y_max * unit),
            f"{z}_min": float(self.z_min * unit),
            f"{z}_max": float(self.z_max * unit),
        }
        sim.addrect(
            name=name,
            material=material_name,
            override_mesh_order_from_material_database=True,
            mesh_order=mesh_order,
            **kwargs,
        )

    def _trimesh(self, color=None, scale=None):
        sx, sy, sz = scale or (1, 1, 1)
        poly = sg.Polygon(
            [
                (self.x_min * sx, self.y_min * sy),
                (self.x_min * sx, self.y_max * sy),
                (self.x_max * sx, self.y_max * sy),
                (self.x_max * sx, self.y_min * sy),
            ],
        )
        prism = extrude_polygon(poly, self.z_max * sz - self.z_min * sz)
        prism = cast(Trimesh, prism.apply_translation((0, 0, self.z_min * sz)))
        if color is not None:
            prism.visual.face_colors = _to_rgba(color)  # type: ignore
        return prism


AxisDirection = Union[Literal["x"], Literal["y"], Literal["z"]]


class Prism(Geometry):
    """A prism is a 2D Polygon extruded along a certain axis direction ('x', 'y', or 'z').

    Note:
        currently only extrusions along 'y' (perpendicular to the chip)
        are fully supported!
    """

    poly: np.ndarray[Tuple[int, Literal[2]], np.dtype[np.float_]] = Field(
        description="the 2D array (Nx2) with polygon vertices"
    )
    h_min: float = Field(description="the start height of the extrusion")
    h_max: float = Field(description="the end height of the extrusion")
    axis: AxisDirection = Field(
        default="y",
        description="the axis along which the polygon will be extruded ('x', 'y', or 'z').",
    )

    def _mask2d_single(self, X, Y, z):
        poly = sg.Polygon(self.poly)
        if self.axis == "x":
            # x, y, z -> y, z, x
            y_min, _ = self.poly.min(0)
            y_max, _ = self.poly.max(0)
            line = sg.LineString([(y_min, z), (y_max, z)])
            intersection = np.asarray(poly.intersection(line).coords)
            if not intersection.shape[0]:
                return np.zeros_like(X, dtype=bool)
            (y_min, _), (y_max, _) = intersection
            y_min, y_max = min(y_min, y_max), max(y_min, y_max)
            x_min, x_max = min(self.h_min, self.h_max), max(self.h_min, self.h_max)
            return (x_min <= X) & (X <= x_max) & (y_min <= Y) & (Y <= y_max)
        elif self.axis == "y":
            # x, y, z -> z, x, y
            _, x_min = self.poly.min(0)
            _, x_max = self.poly.max(0)
            line = sg.LineString([(z, x_min), (z, x_max)])
            intersection = np.asarray(poly.intersection(line).coords)
            if not intersection.shape[0]:
                return np.zeros_like(X, dtype=bool)
            (_, x_min), (_, x_max) = intersection
            x_min, x_max = min(x_min, x_max), max(x_min, x_max)
            y_min, y_max = min(self.h_min, self.h_max), max(self.h_min, self.h_max)
            return (x_min <= X) & (X <= x_max) & (y_min <= Y) & (Y <= y_max)
        else:
            # x, y, z -> x, y, z
            if (z < self.h_min) or (self.h_max < z):
                return np.zeros_like(X, dtype=bool)
            return np.asarray(
                [poly.contains(sg.Point(x, y)) for x, y in zip(X.ravel(), Y.ravel())],
                dtype=bool,
            ).reshape(X.shape)

    def _lumadd(self, sim, material_name, mesh_order, unit=1e-6, xyz="yzx"):
        name = token_hex(4)

        if xyz not in ("xyz", "yzx", "zxy"):
            raise ValueError(
                f"Prism axes should be positively oriented when adding to Lumerical. Got: {xyz!r}"
            )

        sim.addpoly(
            name=name,
            material=material_name,
            override_mesh_order_from_material_database=True,
            mesh_order=mesh_order,
            use_relative_coordinates=True,
            vertices=np.asarray(self.poly, float) * float(unit),
            x=0.0,
            y=0.0,
            z_min=float(self.h_min * unit),
            z_max=float(self.h_max * unit),
        )

        x, y, z = xyz
        if self.axis == "x":
            z, x, y = x, y, z
            # raise NotImplementedError(
            #    "Only prisms extruded perpendicular to the 'chip surface' are currently supported in Lumerical."
            # )
        elif self.axis == "y":
            y, z, x = x, y, z
        xyz = f"{x}{y}{z}"

        if xyz in ["yzx", "zxy"]:
            raise NotImplementedError(
                "Only prisms extruded perpendicular to the 'chip surface' are currently supported in Lumerical."
            )

    def _trimesh(self, color=None, scale=None):
        poly = sg.Polygon(self.poly)
        prism = extrude_polygon(poly, self.h_max - self.h_min)
        prism = prism.apply_translation((0, 0, self.h_min))
        if self.axis == "x":
            prism.vertices = np.roll(prism.vertices, shift=1, axis=1)  # type: ignore
        if self.axis == "y":
            prism.vertices = np.roll(prism.vertices, shift=-1, axis=1)  # type: ignore
        if scale is not None:
            sx, sy, sz = scale
            prism.vertices *= np.array([[sx, sy, sz]])  # type: ignore
        if color is not None:
            prism.visual.face_colors = _to_rgba(color)  # type: ignore
        return prism

    def _center(self):
        import shapely.geometry as sg

        a, b = np.array(sg.Polygon(self.poly).centroid.xy).ravel()
        c = 0.5 * (self.h_min + self.h_max)
        x, y, z = {
            "x": (c, a, b),
            "y": (b, c, a),
            "z": (a, b, c),
        }[self.axis]
        return x, y, z

    def _axis_tuple(self):
        return {
            "x": (1, 0, 0),
            "y": (0, 1, 0),
            "z": (0, 0, 1),
        }[self.axis]


def _to_rgba(c):
    r, g, b, a = _to_rgba_mpl(c)
    a = min(max(a, 0.1), 0.9)
    return float(r), float(g), float(b), float(a)



from scipy.interpolate import RegularGridInterpolator
from scipy.integrate import dblquad, simpson


def integrate_interpolate_2d(x, y, data, extent=None):
    """
    First the data on the given grid is interpolated and then integrated using `scipy.dblquad`.
    This procedure is best suited if one wants to integrate over a (small) region of interest that can be placed at off-grid positions"""
    interp = RegularGridInterpolator((x, y), data)
    if extent is None:
        extent = [[min(x), max(x)], [min(y), max(y)]]

    def integrable(x, y):
        return interp((x, y))

    return dblquad(
        integrable,
        extent[0][0],
        extent[0][1],
        lambda x: extent[1][0],
        lambda x: extent[1][1],
        epsabs=1,
        epsrel=0.001,
    )[0]


def integrate_2d(x, y, data):
    """much simpler integration over the full grid"""
    int1 = simpson(data, y)
    int2 = simpson(int1, x)
    return int2



""" a Structure is a combination of a geometry with a material (and an optional mesh order) """

import numpy as np
from pydantic import Field, validator
from trimesh.scene import Scene
from trimesh.transformations import rotation_matrix

from .base_model import BaseModel
from .geometries import Geometry
from .materials import Material


class Structure(BaseModel):
    """a `Structure` is an association between a `Geometry` and a `Material`"""

    material: Material = Field(description="the material of the structure")
    geometry: Geometry = Field(description="the geometry of the structure")
    mesh_order: int = Field(default=5, description="the mesh order of the structure")

    @validator("material")
    def validate_material(cls, material):
        return Material.parse_obj(material)

    def _lumadd(self, sim, env, unit=1e-6, xyz="yzx"):
        material_name = self.material._lumadd(sim, env, unit)
        self.geometry._lumadd(sim, material_name, self.mesh_order, unit, xyz)

    def _trimesh(self, color=None, scale=None):
        return self.geometry._trimesh(
            color=(color or self.material.meta.get("color")),
            scale=scale,
        )

    def _visualize(self, scale=None):
        return self._trimesh(scale=scale).show()


def visualize_structures(structures, scale=None):
    """easily visualize a collection (list) of `Structure` objects"""
    scene = Scene(
        geometry=[s._trimesh(scale=scale) for s in sort_structures(structures)]
    )
    scene.apply_transform(rotation_matrix(np.pi - np.pi / 6, (0, 1, 0)))
    return scene.show()


def sort_structures(structures):
    struct_info = [(s.mesh_order, -i, s) for i, s in enumerate(structures)]
    sorted_struct_info = sorted(struct_info, key=lambda I: (I[0], I[1]), reverse=True)
    return [s for _, _, s in sorted_struct_info]
